@@ -73,7 +73,7 @@ static unsigned stringToColour(const ::WPXString &s)
   return val;
 }
 
-libvisio::VSDSVGGenerator::VSDSVGGenerator(std::ostream &output_sink): m_gradient(), m_style(), m_gradientIndex(1), m_shadowIndex(1), m_isFirstPage(true), m_outputSink(output_sink)
+libvisio::VSDSVGGenerator::VSDSVGGenerator(libvisio::VSDStringVector &vec): m_gradient(), m_style(), m_gradientIndex(1), m_shadowIndex(1), m_outputSink(), m_vec(vec)
 {
 }
 
@@ -83,17 +83,6 @@ libvisio::VSDSVGGenerator::~VSDSVGGenerator()
 
 void libvisio::VSDSVGGenerator::startGraphics(const WPXPropertyList &propList)
 {
-  if (m_isFirstPage)
-    m_isFirstPage = false;
-  else
-    m_outputSink << "<hr/>\n";
-
-  m_outputSink << "<!-- \n";
-  m_outputSink << "<?xml version=\"1.0\" encoding=\"UTF-8\" standalone=\"no\"?>\n";
-  m_outputSink << "<!DOCTYPE svg PUBLIC \"-//W3C//DTD SVG 1.1//EN\"";
-  m_outputSink << " \"http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd\">\n";
-  m_outputSink << " -->\n";
-
   m_outputSink << "<svg:svg version=\"1.1\" xmlns:svg=\"http://www.w3.org/2000/svg\" xmlns:xlink=\"http://www.w3.org/1999/xlink\" ";
   if (propList["svg:width"])
     m_outputSink << "width=\"" << doubleToString(72*(propList["svg:width"]->getDouble())) << "\" ";
@@ -105,6 +94,8 @@ void libvisio::VSDSVGGenerator::startGraphics(const WPXPropertyList &propList)
 void libvisio::VSDSVGGenerator::endGraphics()
 {
   m_outputSink << "</svg:svg>\n";
+  m_vec.append(m_outputSink.str().c_str());
+  m_outputSink.str("");
 }
 
 void libvisio::VSDSVGGenerator::setStyle(const ::WPXPropertyList &propList, const ::WPXPropertyListVector &gradient)
@@ -330,10 +321,9 @@ void libvisio::VSDSVGGenerator::drawEllipse(const WPXPropertyList &propList)
   m_outputSink << "rx=\"" << doubleToString(72*propList["svg:rx"]->getDouble()) << "\" ry=\"" << doubleToString(72*propList["svg:ry"]->getDouble()) << "\" ";
   writeStyle();
   if (propList["libwpg:rotate"] && propList["libwpg:rotate"]->getDouble() != 0.0)
-    m_outputSink << " transform=\" translate(" << doubleToString(72*propList["svg:cx"]->getDouble()) << ", " << doubleToString(72*propList["svg:cy"]->getDouble())
-                 << ") rotate(" << doubleToString(-propList["libwpg:rotate"]->getDouble())
-                 << ") translate(" << doubleToString(-72*propList["svg:cx"]->getDouble())
-                 << ", " << doubleToString(-72*propList["svg:cy"]->getDouble())
+    m_outputSink << " transform=\" rotate(" << doubleToString(-propList["libwpg:rotate"]->getDouble())
+                 << ", " << doubleToString(72*propList["svg:cy"]->getDouble())
+                 << ", " << doubleToString(72*propList["svg:cy"]->getDouble())
                  << ")\" ";
   m_outputSink << "/>\n";
 }
@@ -469,15 +459,61 @@ void libvisio::VSDSVGGenerator::drawGraphicObject(const ::WPXPropertyList &propL
 
 void libvisio::VSDSVGGenerator::startTextObject(const ::WPXPropertyList &propList, const ::WPXPropertyListVector & /* path */)
 {
+  double x = 0.0;
+  double y = 0.0;
+  double width = 0.0;
+  double height = 0.0;
   m_outputSink << "<svg:text ";
   if (propList["svg:x"] && propList["svg:y"])
-    m_outputSink << "x=\"" << doubleToString(72*(propList["svg:x"]->getDouble())) << "\" y=\"" << doubleToString(72*(propList["svg:y"]->getDouble())) << "\"";
+  {
+    x = propList["svg:x"]->getDouble();
+    y = propList["svg:y"]->getDouble();
+  }
+
+  double xmiddle = x;
+  double ymiddle = y;
+
+  if (propList["svg:width"])
+  {
+    width = propList["svg:width"]->getDouble();
+    xmiddle += width / 2.0;
+  }
+
+  if (propList["svg:height"])
+  {
+    height = propList["svg:height"]->getDouble();
+    ymiddle += height / 2.0;
+  }
+
+  if (propList["draw:textarea-vertical-align"])
+  {
+    if (propList["draw:textarea-vertical-align"]->getStr() == "middle")
+      y = ymiddle;
+    if (propList["draw:textarea-vertical-align"]->getStr() == "bottom")
+    {
+      y += height;
+      if (propList["fo:padding-bottom"])
+        y -= propList["fo:padding-bottom"]->getDouble();
+    }
+  }
+  else
+    y += height;
+
+  if (propList["fo:padding-left"])
+    x += propList["fo:padding-left"]->getDouble();
+
+  m_outputSink << "x=\"" << doubleToString(72*x) << "\" y=\"" << doubleToString(72*y) << "\"";
+
+  // rotation is around the center of the object's bounding box
   if (propList["libwpg:rotate"] && propList["libwpg:rotate"]->getDouble() != 0.0)
-    m_outputSink << " transform=\"translate(" << doubleToString(72*propList["svg:x"]->getDouble()) << ", " << doubleToString(72*propList["svg:y"]->getDouble())
-                 << ") rotate(" << doubleToString(-propList["libwpg:rotate"]->getDouble())
-                 << ") translate(" << doubleToString(-72*propList["svg:x"]->getDouble())
-                 << ", " << doubleToString(-72*propList["svg:y"]->getDouble())
-                 << ")\"";
+  {
+    double angle(propList["libwpg:rotate"]->getDouble());
+    while (angle > 180.0)
+      angle -= 360.0;
+    while (angle < -180.0)
+      angle += 360.0;
+    m_outputSink << " transform=\"rotate(" << doubleToString(angle) << ", " << doubleToString(72*xmiddle) << ", " << doubleToString(72*ymiddle) << ")\" ";
+  }
   m_outputSink << ">\n";
 
 }
