@@ -451,15 +451,119 @@ void libvisio::VSDXContentCollector::_lineProperties(double strokeWidth, Colour
   // Deal with line markers (arrows, etc.)
   if (startMarker > 0)
   {
-    m_styleProps.insert("draw:marker-start-viewbox", "0 0 20 30");
-    m_styleProps.insert("draw:marker-start-path", "m10 0-10 30h20z");
-    m_styleProps.insert("draw:marker-start-width", m_scale*0.118);
+    m_styleProps.insert("draw:marker-start-viewbox", _linePropertiesMarkerViewbox(startMarker));
+    m_styleProps.insert("draw:marker-start-path", _linePropertiesMarkerPath(startMarker));
+    m_styleProps.insert("draw:marker-start-width", m_scale*_linePropertiesMarkerScale(startMarker)*(0.1/(strokeWidth*strokeWidth+1)+2.54*strokeWidth));
   }
-  else if (endMarker > 0)
+  if (endMarker > 0)
   {
-    m_styleProps.insert("draw:marker-end-viewbox", "0 0 20 30");
-    m_styleProps.insert("draw:marker-end-path", "m10 0-10 30h20z");
-    m_styleProps.insert("draw:marker-end-width", m_scale*0.118);
+    m_styleProps.insert("draw:marker-end-viewbox", _linePropertiesMarkerViewbox(endMarker));
+    m_styleProps.insert("draw:marker-end-path", _linePropertiesMarkerPath(endMarker));
+    m_styleProps.insert("draw:marker-end-width", m_scale*_linePropertiesMarkerScale(endMarker)*(0.1/(strokeWidth*strokeWidth+1)+2.54*strokeWidth));
+  }
+}
+
+const char *libvisio::VSDXContentCollector::_linePropertiesMarkerViewbox(unsigned marker)
+{
+  switch (marker)
+  {
+  case 1:
+  case 2:
+  case 9:
+  case 15:
+    return "0 0 20 10";
+  case 8:
+    return "0 0 20 18";
+  case 3:
+  case 4:
+  case 5:
+  case 6:
+  case 11:
+  case 16:
+  case 17:
+  case 18:
+    return "0 0 20 20";
+  case 12:
+  case 13:
+  case 14:
+    return "0 0 20 30";
+  case 22:
+  case 39:
+    return "0 0 20 40";
+  case 21:
+    return "0 0 30 30";
+  case 10:
+    return "0 0 1131 1131";
+  default:
+    return "0 0 20 30";
+  }
+}
+
+const char *libvisio::VSDXContentCollector::_linePropertiesMarkerPath(unsigned marker)
+{
+  switch (marker)
+  {
+  case 1:
+    return "m10 -4l-14 14l4 4l10 -10l10 10l4 -4z";
+  case 2:
+    return "m10 0-10 10h20z";
+  case 3:
+    return "m10 -8l-14 28l6 3l8 -16l8 16l6 -3z";
+  case 4:
+    return "m10 0-10 20h20z";
+  case 5:
+    return "m10 0-10 20q10,-5 20,0z";
+  case 6:
+    return "m10 0-10 20q10,5 20,0z";
+  case 8:
+    return "m10 0q-2.6,13.4 -10,18q10,-5 20,0q-7.4,-4.6 -10,-18";
+  case 9:
+    return "m-2 -8l4 -4l20 20l-4 4z";
+  case 10: // Copied from what LO exports when using the "circle" marker
+    return "m462 1118-102-29-102-51-93-72-72-93-51-102-29-102-13-105 13-102 29-106 51-102 72-89 93-72 102-50 102-34 106-9 101 9 106 34 98 50 93 72 72 89 51 102 29 106 13 102-13 105-29 102-51 102-72 93-93 72-98 51-106 29-101 13z";
+  case 11:
+    return "m0 0v10h10v-10z";
+  case 12:
+    return "m10 -12l-14 42l9 3l5 -15l5 15l9 -3z";
+  case 13:
+    return "m10 0-10 30h20z";
+  case 14:
+    return "m10 0-10 30h20z m-10 -18l-5 15h10z";
+  case 15:
+    return "m10 0-10 10h20z m-10 -7l-5 5h10z";
+  case 16:
+    return "m10 0-10 20h20z m-10 -13l-5 10h10z";
+  case 17:
+    return "m10 0-10 20q10,-5 20,0z m-10 -13l-4 8q4,-2 8,0z";
+  case 18:
+    return "m10 0-10 20q10,5 20,0z m-10 -13l-5 10q5,2 10,0z";
+  case 21:
+    return "m0 0v30h30v-30z m-20 10v10h10v-10z";
+  case 22:
+    return "m10 0-10 20l10 20l10 -20z m-10 -12l-6 12l6 12l6 -12z";
+  case 39:
+    return "m10 0-10 20h20z m-10 0-10 20h20z";
+  default:
+    return "m10 0-10 30h20z";
+  }
+}
+
+double libvisio::VSDXContentCollector::_linePropertiesMarkerScale(unsigned marker)
+{
+  switch (marker)
+  {
+  case 11:
+  case 10:
+    return 0.7;
+  case 14:
+  case 15:
+  case 16:
+  case 17:
+  case 18:
+  case 22:
+    return 1.2;
+  default:
+    return 1.0;
   }
 }
 
@@ -486,7 +590,10 @@ void libvisio::VSDXContentCollector::_flushCurrentPath()
     }
     else if (m_currentGeometry[i]["libwpg:path-action"]->getStr() == "M")
     {
-      if ((startX == x && startY == y) && path.count() && !wasMove)
+      if (((startX == x && startY == y)
+           || (m_styleProps["draw:fill"] && m_styleProps["draw:fill"]->getStr() != "none"
+               && m_styleProps["draw:stroke"] && m_styleProps["draw:stroke"]->getStr() == "none"))
+          && path.count() && !wasMove)
       {
         WPXPropertyList closedPath;
         closedPath.insert("libwpg:path-action", "Z");
@@ -506,7 +613,10 @@ void libvisio::VSDXContentCollector::_flushCurrentPath()
     }
     path.append(m_currentGeometry[i]);
   }
-  if ((startX == x && startY == y) && path.count() && !wasMove)
+  if (((startX == x && startY == y)
+       || (m_styleProps["draw:fill"] && m_styleProps["draw:fill"]->getStr() != "none"
+           && m_styleProps["draw:stroke"] && m_styleProps["draw:stroke"]->getStr() == "none"))
+      && path.count() && !wasMove)
   {
     WPXPropertyList closedPath;
     closedPath.insert("libwpg:path-action", "Z");
@@ -523,15 +633,32 @@ void libvisio::VSDXContentCollector::_flushCurrentPath()
 void libvisio::VSDXContentCollector::_flushText()
 {
   if (!m_textStream.size()) return;
+
+  double xmiddle = m_txtxform ? m_txtxform->width / 2.0 : m_xform.width / 2.0;
+  double ymiddle = m_txtxform ? m_txtxform->height / 2.0 : m_xform.height / 2.0;
+
+  transformPoint(xmiddle,ymiddle, m_txtxform);
+
+  double x = xmiddle - (m_txtxform ? m_txtxform->width / 2.0 : m_xform.width / 2.0);
+  double y = ymiddle - (m_txtxform ? m_txtxform->height / 2.0 : m_xform.height / 2.0);
+
   double angle = 0.0;
   transformAngle(angle, m_txtxform);
 
-  double x = 0.0;
-  double y = m_txtxform ? m_txtxform->height: m_xform.height;
-
-  transformPoint(x,y, m_txtxform);
-
   WPXPropertyList textBlockProps;
+
+  bool flipX = false;
+  bool flipY = false;
+  transformFlips(flipX, flipY);
+
+  if (flipX)
+    angle -= M_PI;
+
+  while (angle > M_PI)
+    angle -= 2 * M_PI;
+  while (angle < -M_PI)
+    angle += 2 * M_PI;
+
   textBlockProps.insert("svg:x", m_scale * x);
   textBlockProps.insert("svg:y", m_scale * y);
   textBlockProps.insert("svg:height", m_scale * (m_txtxform ? m_txtxform->height : m_xform.height));
@@ -540,7 +667,7 @@ void libvisio::VSDXContentCollector::_flushText()
   textBlockProps.insert("fo:padding-bottom", m_textBlockStyle.bottomMargin);
   textBlockProps.insert("fo:padding-left", m_textBlockStyle.leftMargin);
   textBlockProps.insert("fo:padding-right", m_textBlockStyle.rightMargin);
-  textBlockProps.insert("libwpg:rotate", -angle*180/M_PI, WPX_GENERIC);
+  textBlockProps.insert("libwpg:rotate", angle*180/M_PI, WPX_GENERIC);
 
   switch (m_textBlockStyle.verticalAlign)
   {
@@ -725,92 +852,15 @@ void libvisio::VSDXContentCollector::_flushText()
 
 void libvisio::VSDXContentCollector::_flushCurrentForeignData()
 {
-  double x1 = m_foreignOffsetX;
-  double y1 = m_foreignOffsetY;
-  double x2 = m_foreignOffsetX + m_foreignWidth;
-  double y2 = m_foreignOffsetY + m_foreignHeight;
-  double x3 = x1;
-  double y3 = y2;
-  double x4 = x2;
-  double y4 = y1;
-  transformPoint(x1,y1);
-  transformPoint(x2,y2);
-  transformPoint(x3,y3);
-  transformPoint(x4,y4);
-  double angle = 0.0;
-  transformAngle(angle);
+  double xmiddle = m_foreignOffsetX + m_foreignWidth / 2.0;
+  double ymiddle = m_foreignOffsetY + m_foreignHeight / 2.0;
 
-  double xmin = x1;
-  double xmax = x1;
-
-  if (xmin > x2)
-    xmin = x2;
-  if (xmax < x2)
-    xmax = x2;
-
-  if (xmin > x3)
-    xmin = x3;
-  if (xmax < x3)
-    xmax = x3;
-
-  if (xmin > x4)
-    xmin = x4;
-  if (xmax < x4)
-    xmax = x4;
-
-  double ymin = y1;
-  double ymax = y1;
-
-  if (ymin > y2)
-    ymin = y2;
-  if (ymax < y2)
-    ymax = y2;
-
-  if (ymin > y3)
-    ymin = y3;
-  if (ymax < y3)
-    ymax = y3;
-
-  if (ymin > y4)
-    ymin = y4;
-  if (ymax < y4)
-    ymax = y4;
-
-  double xmiddle = (xmax + xmin) / 2.0;
-  double ymiddle = (ymax + ymin) / 2.0;
+  transformPoint(xmiddle, ymiddle);
 
   bool flipX = false;
   bool flipY = false;
 
-  // If any parent group is flipped, invert flips and
-  // revert the flip information that is already incorporated
-  // in the angle.
-  unsigned shapeId = m_currentShapeId;
-  while (true)
-  {
-    std::map<unsigned, XForm>::iterator iterX = m_groupXForms.find(shapeId);
-    if (iterX != m_groupXForms.end())
-    {
-      XForm xform = iterX->second;
-      if (xform.flipX)
-      {
-        flipX = !flipX;
-        angle = M_PI - angle;
-      }
-      if (xform.flipY)
-      {
-        flipY = !flipY;
-        angle *= -1.0;
-      }
-    }
-    else
-      break;
-    std::map<unsigned, unsigned>::iterator iter = m_groupMemberships.find(shapeId);
-    if (iter != m_groupMemberships.end())
-      shapeId = iter->second;
-    else
-      break;
-  }
+  transformFlips(flipX, flipY);
 
   WPXPropertyList styleProps(m_styleProps);
 
@@ -819,10 +869,19 @@ void libvisio::VSDXContentCollector::_flushCurrentForeignData()
   m_currentForeignProps.insert("svg:y", m_scale*(ymiddle - (m_foreignHeight / 2.0)));
   m_currentForeignProps.insert("svg:height", m_scale*m_foreignHeight);
 
+  double angle = 0.0;
+  transformAngle(angle);
+
   if (flipX)
+  {
     m_currentForeignProps.insert("draw:mirror-horizontal", true);
+    angle = M_PI - angle;
+  }
   if (flipY)
+  {
     m_currentForeignProps.insert("draw:mirror-vertical", true);
+    angle *= -1.0;
+  }
 
   if (angle != 0.0)
     m_currentForeignProps.insert("libwpg:rotate", angle * 180 / M_PI, WPX_GENERIC);
@@ -1208,6 +1267,7 @@ void libvisio::VSDXContentCollector::_handleForeignData(const WPXBinaryData &bin
   else if (m_foreignType == 2)
   {
     m_currentForeignProps.insert("libwpg:mime-type", "object/ole");
+    m_currentForeignData.append(binaryData);
   }
 }
 
@@ -1301,27 +1361,9 @@ void libvisio::VSDXContentCollector::collectArcTo(unsigned /* id */, unsigned le
     double chord = sqrt(pow((y2 - m_y),2) + pow((x2 - m_x),2));
     double radius = (4 * bow * bow + chord * chord) / (8 * fabs(bow));
     int largeArc = fabs(bow) > radius ? 1 : 0;
-    int sweep = bow < 0 ? 1 : 0;
+    bool sweep = (bow < 0);
+    transformFlips(sweep, sweep);
 
-    // If any parent group is flipped, invert sweep
-    unsigned shapeId = m_currentShapeId;
-    while (true)
-    {
-      std::map<unsigned, XForm>::iterator iterX = m_groupXForms.find(shapeId);
-      if (iterX != m_groupXForms.end())
-      {
-        XForm xform = iterX->second;
-        if (xform.flipX) sweep = sweep == 0 ? 1 : 0;
-        if (xform.flipY) sweep = sweep == 0 ? 1 : 0;
-      }
-      else
-        break;
-      std::map<unsigned, unsigned>::iterator iter = m_groupMemberships.find(shapeId);
-      if (iter != m_groupMemberships.end())
-        shapeId = iter->second;
-      else
-        break;
-    }
     m_x = x2;
     m_y = y2;
     arc.insert("svg:rx", m_scale*radius);
@@ -1374,7 +1416,7 @@ void libvisio::VSDXContentCollector::collectNURBSTo(unsigned /* id */, unsigned
     NURBS.insert("libwpg:path-action", "L");
     double nextX = 0;
     double nextY = 0;
-    double denominator = 0.0000001;
+    double denominator = 1E-10;
 
     for (unsigned p = 0; p < controlPoints.size() && p < weights.size(); p++)
     {
@@ -1396,6 +1438,12 @@ void libvisio::VSDXContentCollector::collectNURBSTo(unsigned /* id */, unsigned
   m_x = x2;
   m_y = y2;
   transformPoint(m_x, m_y);
+  NURBS.clear();
+  NURBS.insert("libwpg:path-action", "L");
+  NURBS.insert("svg:x", m_scale*m_x);
+  NURBS.insert("svg:y", m_scale*m_y);
+  m_currentGeometry.push_back(NURBS);
+
 }
 
 double libvisio::VSDXContentCollector::_NURBSBasis(unsigned knot, unsigned degree, double point, const std::vector<double> &knotVector)
@@ -1636,6 +1684,37 @@ void libvisio::VSDXContentCollector::transformAngle(double &angle, XForm *txtxfo
   angle = fmod(2.0*M_PI + (y1 > y0 ? 1.0 : -1.0)*acos((x1-x0) / sqrt((x1-x0)*(x1-x0) + (y1-y0)*(y1-y0))), 2.0*M_PI);
 }
 
+void libvisio::VSDXContentCollector::transformFlips(bool &flipX, bool &flipY)
+{
+  if (!m_isShapeStarted)
+    return;
+
+  if (!m_currentShapeId)
+    return;
+
+  unsigned shapeId = m_currentShapeId;
+
+  while (true)
+  {
+    std::map<unsigned, XForm>::iterator iterX = m_groupXForms.find(shapeId);
+    if (iterX != m_groupXForms.end())
+    {
+      XForm xform = iterX->second;
+      if (xform.flipX)
+        flipX = !flipX;
+      if (xform.flipY)
+        flipY = !flipY;
+    }
+    else
+      break;
+    std::map<unsigned, unsigned>::iterator iter = m_groupMemberships.find(shapeId);
+    if (iter != m_groupMemberships.end())
+      shapeId = iter->second;
+    else
+      break;
+  }
+}
+
 void libvisio::VSDXContentCollector::collectShapeId(unsigned /* id */, unsigned level, unsigned /* shapeId */)
 {
   _handleLevelChange(level);
@@ -1751,6 +1830,7 @@ void libvisio::VSDXContentCollector::collectShape(unsigned id, unsigned level, u
         m_foreignOffsetY = m_stencilShape->m_foreign->offsetY;
         m_foreignWidth = m_stencilShape->m_foreign->width;
         m_foreignHeight = m_stencilShape->m_foreign->height;
+        m_currentForeignData.clear();
         _handleForeignData(m_stencilShape->m_foreign->data);
       }
 
